import types
from library_class import LibraryClass
from decorules.decorators import raise_if_false_on_class, raise_if_false_on_instance, enforces_instance_rules

is_multiplier_larger_than_1 = lambda x: x.MULTIPLIER>1.0
smallest_coordinate_times_multiplier_larger_than_largest = lambda x: min(x.coordinates)*x.MULTIPLIER>max(x.coordinates)
@raise_if_false_on_instance(smallest_coordinate_times_multiplier_larger_than_largest,
                            ValueError,
                            "Checks that the multiplier times the smallest coordinate is larger than the largest coordinate")
@raise_if_false_on_class(is_multiplier_larger_than_1,
                         AttributeError,
                         "Checks that the multiplier is larger than 2.0")
class ClientClass(LibraryClass):
    MULTIPLIER = 2.5
    @enforces_instance_rules
    def append(self, value=0.0):
        self.coordinates.append(value)
    pass

def main():
    a = LibraryClass("a")
    a.testme()
    b = LibraryClass("b", 0.2, 0.5)
    b.testme()
    try:
        c = LibraryClass("c", -2.0, 0.5)
        c.testme()
    except ValueError as ve:
        print(ve)
        pass
    d = ClientClass("d", 0.1,0.1,0.05)
    d.testme()
    try:
        e = ClientClass("e",0.1,0.1,0.025)
        e.testme()
    except ValueError as ve:
        print(ve)
        pass
    try:
        f = ClientClass("f",0.8,0.8,0.4)
        f.testme()
    except ValueError as ve:
        print(ve)
        pass
    g = ClientClass("g", 0.1,0.1,0.05)
    g.testme()
    g.append(0.06)  # should not break an instance rule
    g.testme()
    try:
        g.append(max(g.coordinates)*ClientClass.MULTIPLIER) # should break
    except ValueError as ve:
        print(ve)
        pass


if __name__ == "__main__":
    main()




import types
from functools import partial
from decorules.has_enforced_rules import HasEnforcedRules
from decorules.decorators import raise_if_false_on_class, raise_if_false_on_instance
from decorules.predicates import key_type_enforcer
from collections.abc import Iterable


def key_structure_enforcer(input_type,
                      enforced_type: type,
                      enforced_key: str):
    member_object = getattr(input_type, enforced_key, None)
    if member_object is None:
        return False
    else:
        return issubclass(type(member_object), enforced_type)
    pass


has_static_multiplier = partial(key_type_enforcer, enforced_type=float, enforced_key='MULTIPLIER', attrs=None)
has_testme_method = partial(key_type_enforcer, enforced_type=types.FunctionType, enforced_key='testme', attrs=None)
are_coordinates_set_after_init = partial(key_type_enforcer, enforced_type=Iterable, enforced_key='coordinates')
are_coordinates_within_distance_1 = lambda y: (sum([x ** 2 for x in y.coordinates]) ** 0.5)<=1.0
has_second_static_multiplier = partial(key_structure_enforcer, enforced_type=float, enforced_key='NOT_USED_MULTIPLIER')

@raise_if_false_on_instance(are_coordinates_within_distance_1, ValueError, 'initial coordinates are outside the unit circle (euclidean distance)')
@raise_if_false_on_instance(are_coordinates_set_after_init, AttributeError, 'Missing coordinates after initialization')
@raise_if_false_on_class(has_testme_method, AttributeError, "Checks if a method called testme is present")
@raise_if_false_on_class(has_static_multiplier, AttributeError, "Checks if a float multiplier is set on the class")
class LibraryClass(metaclass=HasEnforcedRules):
    MULTIPLIER = 0.5
    NOT_USED_MULTIPLIER = 20.0
    def __init__(self, name : str, *args):
        self.name = name
        self.coordinates = list(args)
    def testme(self):
        print(self.name)
    pass






